using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;

namespace SerialPix
{
    public partial class frmMain : Form
    {
        //function pointer so progress bar can be updated from serialport callback thread
        public delegate void ProgressDelegate(int val);

        //imported function for locking display during updates (clean up refresh/resize)
        [System.Runtime.InteropServices.DllImport("user32.dll")]
        public static extern bool LockWindowUpdate(IntPtr hWndLock);

        private long[] commrates = { 110, 300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600};
        private string[] commports;

        //default comm. rate = 57.6K
        private int default_commrate = 9;
        //default comm. port = 4 (or highest)
        private int default_commport = 3;

        private string commrate;
        private string commport;

        //display config variables... display preview colors? Stretch picture? black out right side?
        private bool isPreview = false;
        private bool isStretch = true;
        private bool doBorder = true;

        private int displayWidth = 256;
        private int displayHeight = 224;

        private byte startByte = 0x7F; //make all almost-whites into whites
        private byte startByteReplace = 0xFF;

        private byte stopByte = 0x01; //and all almost-blacks into blacks
        private byte stopByteReplace = 0x00;

        //internal stored original and sendable bitmaps
        private Bitmap full_bmp, reduced_bmp;

        //set to true when cancelled to allow loop to exit
        private bool cancelFlag;

        //constructor for form (at class build)
        public frmMain()
        {
            InitializeComponent();
        }

        //initialize windows form components (some can only be done at runtime)
        private void frmMain_Load(object sender, EventArgs e)
        {
            commports = SerialPort.GetPortNames();
            default_commport = Math.Min(default_commport, commports.Length - 1);
            //fill comm port menu and check default
            foreach(string s in commports)
                commPortToolStripMenuItem.DropDownItems.Add(s, null, commPortMenuItem_Click);

            ToolStripMenuItem mnuTemp = (ToolStripMenuItem)(commPortToolStripMenuItem.DropDownItems[default_commport]);
            mnuTemp.Checked = true;
            commport = commports[default_commport];

            //fill comm speed menu and check default
            foreach (long l in commrates)
                commRateToolStripMenuItem.DropDownItems.Add(l.ToString(), null, commSpeedMenuItem_Click);

            mnuTemp = (ToolStripMenuItem)(commRateToolStripMenuItem.DropDownItems[default_commrate]);
            mnuTemp.Checked = true;
            commrate = commrates[default_commrate].ToString();

            stretchImageToolStripMenuItem.Checked = isStretch;
            cropImageToolStripMenuItem.Checked = !isStretch;

            cancelTransmissionToolStripMenuItem.Enabled = false;

            //init bitmaps so they are never null (could be a problem later)
            full_bmp = new Bitmap(1, 1);
            reduced_bmp = new Bitmap(1, 1);

            frmMain_Resize(sender, e);

            lblStatus.Font = SystemFonts.StatusFont;
            lblStatus.Left = progressBar.Left + 3;
            lblStatus.Top = progressBar.Top + 3;

            //init serial port
            serialPort.DataReceived += new SerialDataReceivedEventHandler(serialPort_DataReceived);
            serialPort.PortName = commport;
            serialPort.BaudRate = int.Parse(commrate);
        }

        //resize form components to fit within new window size
        //lockwindowupdate just helps avoid staggered resizing of controls
        void frmMain_Resize(object sender, System.EventArgs e)
        {
            LockWindowUpdate(this.Handle);

            this.Width = (this.Width - this.ClientSize.Width) + displayWidth + 2;
            this.Height = (this.Height - this.ClientSize.Height) + displayHeight + mnuMain.Height + progressBar.Height + 3;

            lblStatus.Left = progressBar.Left + 3;
            lblStatus.Top = progressBar.Top + 3;

            LockWindowUpdate(IntPtr.Zero);
            this.Refresh();
        }

        //the form's repaint event
        //reduceBitmap will redraw the original bitmap into reduced_bmp if it doesn't already
        //  contain the right image/size
        //since reduced_bmp is already the correct size, just draw it unscaled
        void frmMain_Paint(object sender, System.Windows.Forms.PaintEventArgs e)
        {
            reduceBitmap();
            e.Graphics.DrawImageUnscaled(reduced_bmp, 1, mnuMain.Height + 1);
            e.Graphics.DrawRectangle(Pens.Black, 0, mnuMain.Height, this.ClientSize.Width - 1, this.ClientSize.Height - mnuMain.Height - progressBar.Height - 3);
        }

        //enable the drag/drop option when a drag enters the form
        void frmMain_DragEnter(object sender, System.Windows.Forms.DragEventArgs e)
        {
            e.Effect = DragDropEffects.Move;
        }

        //load the dragged filename
        void frmMain_DragDrop(object sender, System.Windows.Forms.DragEventArgs e)
        {
            if (!e.Data.GetDataPresent(DataFormats.FileDrop))
                return;

            object o = e.Data.GetData(DataFormats.FileDrop);
            string filename = (string)(((System.Array)o).GetValue(0));

            loadImage(filename);
        }

        //bring up load file dialog and.... load the file
        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            DialogResult result = dlgOpen.ShowDialog();
            if (result == DialogResult.OK)
                loadImage(dlgOpen.FileName);
        }

        //show what the image would look like flattened
        private void previewColorsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            isPreview = true;
            this.Refresh();
        }

        //initiate sent by sending startbyte and waiting for response
        private void sendImageToolStripMenuItem_Click(object sender, EventArgs e)
        {
            startSendBitmap();
        }

        //tell program that the image should be stretched to current resolution
        private void stretchImageToolStripMenuItem_Click(object sender, EventArgs e)
        {
            stretchImageToolStripMenuItem.Checked = true;
            cropImageToolStripMenuItem.Checked = false;

            isStretch = stretchImageToolStripMenuItem.Checked;
            this.Refresh();
        }

        //tell program that the image should be cropped instead of stretched
        private void cropImageToolStripMenuItem_Click(object sender, EventArgs e)
        {
            stretchImageToolStripMenuItem.Checked = false;
            cropImageToolStripMenuItem.Checked = true;

            isStretch = stretchImageToolStripMenuItem.Checked;
            this.Refresh();
        }

        //bring up a custom dialog window to allow setting the horiz/vert resolution
        private void setResolutionToolStripMenuItem_Click(object sender, EventArgs e)
        {
            frmRes f = new frmRes(displayWidth, displayHeight);
            f.ShowDialog();

            if (f.DialogResult == DialogResult.Cancel)
                return;

            int[] results = (int[])(f.Tag);
            displayWidth = results[0];
            displayHeight = results[1];

            frmMain_Resize(null, null);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        //select the comm. rate to use
        private void commSpeedMenuItem_Click(object sender, EventArgs e)
        {
            if (serialPort.IsOpen)
            {
                ErrorBox("Please wait until transfer is complete.");
                return;
            }

            ToolStripMenuItem mnuSender = (ToolStripMenuItem)sender;
            foreach (ToolStripMenuItem t in mnuSender.GetCurrentParent().Items)
                t.Checked = false;
            mnuSender.Checked = true;

            commrate = mnuSender.Text;
            serialPort.BaudRate = int.Parse(commrate);
        }

        //select comm port to use
        private void commPortMenuItem_Click(object sender, EventArgs e)
        {
            if (serialPort.IsOpen)
            {
                ErrorBox("Please wait until transfer is complete.");
                return;
            }

            ToolStripMenuItem mnuSender = (ToolStripMenuItem)sender;
            foreach (ToolStripMenuItem t in mnuSender.GetCurrentParent().Items)
                t.Checked = false;
            mnuSender.Checked = true;

            commport = mnuSender.Text;
            serialPort.PortName = commport;
        }

        //reverse the bits in the current byte to conform to the current wiring
        private byte reverseByte(byte b)
        {
            return (byte)(((b << 7) & 0x80) | ((b << 5) & 0x40) | ((b << 3) & 0x20) | ((b << 1) & 0x10) 
                | ((b >> 7) & 0x01) | ((b >> 5) & 0x02) | ((b >> 3) & 0x04) | ((b >> 1) & 0x08));
        }

        //truncate the R,G, and B, concatenate top 3,3,2 bits to get 1byte/pixel
        // if the byte value equals the stop byte (almost black), make it black
        private byte RGBtoByte(Color c)
        {
            byte b = reverseByte((byte)((c.R & 0xE0) | ((c.G & 0xE0) >> 3) | ((c.B & 0xC0) >> 6)));
            if (b == stopByte)
                b = stopByteReplace;

            return b;
        }

        //reverse the previous function by making full R,G,B values for flattening preview
        private Color BytetoRGB(byte b)
        {
            byte b2 = reverseByte(b);
            return Color.FromArgb(
                (int)(255 * ((b2 >> 5) & 0x07) / 7.0),
                (int)(255 * ((b2 >> 2) & 0x07) / 7.0),
                (int)(255 * ((b2 & 0x03) / 3.0)));
        }

        //load the image from a file... does all the necessary error checks
        private void loadImage(string filename)
        {
            Bitmap new_bmp = null;

            try
            {
                new_bmp = (Bitmap)(Image.FromFile(filename));
            }
            catch (OutOfMemoryException ex)
            {
                ErrorBox("Unrecognized image format.");
                return;
            }
            catch (FileNotFoundException ex)
            {
                ErrorBox("File not found.");
                return;
            }
            catch (Exception ex)
            {
                ErrorBox("An error occurred when loading the file.");
                return;
            }
            full_bmp.Dispose();

            full_bmp = new_bmp;

            
            this.Refresh();
        }

        //stretch/crop image if necessary
        // if trying to preview flattened image: flatten,reverse, and display
        private void reduceBitmap()
        {
            reduced_bmp.Dispose();
            reduced_bmp = new Bitmap(displayWidth, displayHeight);

            Graphics g = Graphics.FromImage(reduced_bmp);

            g.Clear(Color.Gray);
            if (isStretch)
                g.DrawImage(full_bmp, new Rectangle(0, 0, displayWidth, displayHeight),
                    new Rectangle(0, 0, full_bmp.Width-1, full_bmp.Height-1), GraphicsUnit.Pixel);
            //crop image to fit 255x224
            else
                g.DrawImageUnscaled(full_bmp, 0, 0, displayWidth, displayHeight);

            if (isPreview)
            {
                previewColors_unsafe();
                //previewColors();
                isPreview = false;
            }
        }

        //flatten/reverse via getpixel (slow)
        private void previewColors()
        {
            progressBar.Value = 0;
            lblStatus.Visible = false;
            for (int i = 0; i < reduced_bmp.Width; i++)
            {
                for (int j = 0; j < reduced_bmp.Height; j++)
                    reduced_bmp.SetPixel(i, j, BytetoRGB(RGBtoByte(reduced_bmp.GetPixel(i, j))));
                progressBar.Value = (int)(100.0 * i / reduced_bmp.Width);
            }

            progressBar.Value = 0;
            lblStatus.Visible = true;
        }

        //flatten/reverse via unsafe mem ptr loop (fast)
        //thanks to Christian Graus for the info
        //http://www.codeproject.com/cs/media/csharpgraphicfilters11.asp
        private void previewColors_unsafe()
        {
            progressBar.Value = 0;
            lblStatus.Visible = false;

            BitmapData bdata = reduced_bmp.LockBits(new Rectangle(0, 0, displayWidth, displayHeight), ImageLockMode.ReadWrite, PixelFormat.Format32bppRgb);
            int stride = bdata.Stride;
            IntPtr Scan0 = bdata.Scan0;
            Color c;

            unsafe
            {
                byte* p = (byte*)(void*)Scan0;

                int nOffset = stride - reduced_bmp.Width * 4;

                byte red, green, blue;

                for (int j = 0; j < displayHeight; j++)
                {
                    for (int i = 0; i < displayWidth; i++)
                    {
                        blue = p[0];
                        green = p[1];
                        red = p[2];

                        c = BytetoRGB(RGBtoByte(Color.FromArgb(red, green, blue)));

                        p[0] = (byte)(c.B);
                        p[1] = (byte)(c.G);
                        p[2] = (byte)(c.R);

                        p += 4;
                    }
                    p += nOffset;
                    progressBar.Value = (int)(100.0 * j / reduced_bmp.Height);
                }
            }

            reduced_bmp.UnlockBits(bdata);
            progressBar.Value = 0;
            lblStatus.Visible = true;
        }

        //function for testing basic serial functionality
        private void testSend()
        {
            byte[] buffer = new byte[256];
            for (int i = 0; i < buffer.Length; i++)
                buffer[i] = (byte)i;
            serialPort.Open();
            serialPort.Write(buffer, 0, buffer.Length);
            serialPort.Close();
        }

        //event raised when data ping is received... initiate full bitmap send
        void serialPort_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            Console.WriteLine(serialPort.ReadByte().ToString());
            sendBitmap_unsafe(reduced_bmp);
            serialPort.Close();
        }

        //function that updates the progress bar when called from the other thread
        private void updateProgress(int val)
        {
            progressBar.Value = val;
        }

        //initiate send by opening serial port and sending start byte to alert microcontroller
        //byte recognized during vsync on atmel, which sends a byte back and then we can send the
        //whole picture 
        private void startSendBitmap()
        {
            if (serialPort.IsOpen)
            {
                serialPort.Close();
                //ErrorBox("Please wait until transfer is complete.");
                //return;
            }

            cancelFlag = false;
            cancelTransmissionToolStripMenuItem.Enabled = true;

            lblStatus.Visible = false;
            progressBar.Value = 0;

            serialPort.Open();
            serialPort.Write(new byte[] { startByte }, 0, 1);
        }

        //this function is called when the other thread knows it is done sending
        //or when you cancel the xmit
        private void sendCompleted()
        {
            cancelTransmissionToolStripMenuItem.Enabled = false;
            progressBar.Value = 0;
            lblStatus.Visible = true;
        }

        //this send function uses getpixel (slow, but not the bottleneck)
        //loops through, flattens, copies pixels into buffer/sends 1 line at a time
        // and updates progress bar (via delegate since it's in another thread)
        //this function is called from the SerialPort DataReceived thread
        private void sendBitmap(Bitmap bmp)
        {
            int i, j;

            byte[] buffer = new byte[bmp.Width];

            for (j = 0; j < bmp.Height && !cancelFlag; j++)
            {
                for (i = 0; i < bmp.Width; i++)
                {
                    buffer[i] = RGBtoByte(bmp.GetPixel(i, j));
                }
                serialPort.Write(buffer, 0, bmp.Width);
                this.Invoke(new ProgressDelegate(updateProgress), (int)(100.0 * j / reduced_bmp.Height));
            }

            this.Invoke(new MethodInvoker(sendCompleted));
            serialPort.Close();

        }

        //this send function uses unsafe mem pointer (fast, but not the bottleneck anyway)
        //loops through, flattens, copies pixels into buffer/sends 1 line at a time
        // and updates progress bar (via delegate since it's in another thread)
        //this function is called from the SerialPort DataReceived thread
        private void sendBitmap_unsafe(Bitmap bmp_orig)
        {
            Bitmap bmp = (Bitmap)(bmp_orig.Clone());
            BitmapData bdata = bmp.LockBits(new Rectangle(0, 0, bmp.Width, bmp.Height), ImageLockMode.ReadWrite, PixelFormat.Format32bppRgb);
            int stride = bdata.Stride;
            IntPtr Scan0 = bdata.Scan0;
            Color c;
            byte[] buffer = new byte[bmp.Width];
            buffer.Initialize();

            unsafe
            {
                byte* p = (byte*)(void*)Scan0;

                int nOffset = stride - bmp.Width * 4;

                byte red, green, blue;

                for (int j = 0; j < 10; j++)
                {
                    for (int i = 0; i < bmp.Width; i++)
                        buffer[i] = RGBtoByte(Color.FromArgb(0, 0, 0));
                    serialPort.Write(buffer, 0, bmp.Width);
                }

                for (int j = 0; j < bmp.Height && !cancelFlag; j++)
                {
                    for (int i = 0; i < bmp.Width; i++)
                    {
                        blue = p[0];
                        green = p[1];
                        red = p[2];

                        if(!doBorder || i < bmp.Width - 2)
                            buffer[i] = RGBtoByte(Color.FromArgb(red, green, blue));

                        p += 4;
                    }
                    p += nOffset;
                    serialPort.Write(buffer, 0, bmp.Width);
                    this.Invoke(new ProgressDelegate(updateProgress), (int)(100.0 * j / bmp.Height));
                }
            }

            bmp.UnlockBits(bdata);

            //if (cancelFlag)
                serialPort.Write(new byte[] { 0x01 },0,1);

            this.Invoke(new MethodInvoker(sendCompleted));
        }

        //Brings up a custom error message
        private void ErrorBox(string msg)
        {
            MessageBox.Show(msg, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        //paste an image from the windows clipboard
        private void pasteImageToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (Clipboard.ContainsImage())
            {
                full_bmp.Dispose();

                full_bmp = (Bitmap)(Clipboard.GetImage());
                this.Refresh();
            }
        }

        //Show the version of the .NET common-language runtime this computer is running
        private void getNETVersionToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            MessageBox.Show(Environment.Version.ToString(), ".NET Version Info");
        }

        //set the cancel flag so the send loop can exit and execute code to disable menu option, etc...
        private void cancelTransmissionToolStripMenuItem_Click(object sender, EventArgs e)
        {
            cancelFlag = true;
            sendCompleted();
        }
    }
}